#ifndef __SHAPE_ANIMATION_H__
#define __SHAPE_ANIMATION_H__

#include "ShapeModel.h"

class Scene;
class FileSequence;

/**
 * An animating shape.
 */
class ShapeAnimation {
public:
	ShapeAnimation();
	~ShapeAnimation();

	void clear();														/**< Clear all data associated with this animation. */

	//////////////////////////////////////////////////////////////////////////
	// Loading frame data
	//////////////////////////////////////////////////////////////////////////
	void load( const QString &baseDir, const FileSequence &fs, const Scene &scene );								/**< Load a frame sequence. */
	void add( const QString &baseDir, const QString &filename, const Scene &scene, int processFrames );				/**< Add a frame. */
	void add( ShapeModel *shape );																					/**< Add a shape as a frame. */
	void replace ( ShapeModel *shape, int frameIndex );

	//////////////////////////////////////////////////////////////////////////
	// Accessing frame data
	//////////////////////////////////////////////////////////////////////////
	bool validFrame( int frameIndex ) const;							/**< Test whether the index is a valid frame index. */
	size_type numFrames() const;										/**< Number of frames. */
	int frameNumber( const ShapeModel *mesh ) const;					/**< Extract the frame number, given a pointer. */
	const ShapeModel* operator[]( index_type frame ) const;				/**< Constant pointer to a single frame. */
	ShapeModel* operator[]( index_type frame );							/**< Reference to a single frame. */
	ShapeModel* operator[]( const QString &filename );					/**< Reference to a single frame. */


	//////////////////////////////////////////////////////////////////////////
	// Drawing frames
	//////////////////////////////////////////////////////////////////////////
	void draw( index_type frame, bool showBox = false ) const;
	void drawBox( const ShapeModel &model ) const;

private:
	std::vector<ShapeModel*> data;										/**< Shape data. */
};

#endif
