#ifndef __BOOST_GRAPH_DEF_H__
#define __BOOST_GRAPH_DEF_H__

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include "types.h"
#include "PointList.h"

typedef real_type BoostEdgeWeightType;
typedef boost::adjacency_list< boost::vecS, boost::vecS, boost::undirectedS, boost::no_property, boost::property<boost::edge_weight_t, BoostEdgeWeightType> > BoostGraph;
typedef boost::property_map<BoostGraph, boost::vertex_index_t>::type			BoostIndexMap;
typedef boost::property_map<BoostGraph, boost::edge_weight_t>::const_type		BoostEdgeWeightMap_const;
typedef boost::graph_traits<BoostGraph>::vertex_descriptor						BoostVertex;
typedef boost::graph_traits<BoostGraph>::vertex_iterator						BoostVertexIterator;
typedef boost::graph_traits<BoostGraph>::edge_descriptor						BoostEdge;
typedef boost::graph_traits<BoostGraph>::edge_iterator							BoostEdgeIterator;
typedef boost::graph_traits<BoostGraph>::out_edge_iterator						BoostOutEdgeIterator;
typedef boost::graph_traits<BoostGraph>::adjacency_iterator						BoostAdjacencyIterator;
typedef std::vector< std::vector<BoostEdgeWeightType> >							BoostGraphDistance;

// Forward delcarations
namespace Deform {
	namespace DeformAll_Utility {
		struct IndexedWeight;
		typedef std::vector<IndexedWeight> SparseWeight;
		typedef std::deque<SparseWeight> SparseWeightList;
	}
}

// Useful functions
void makeEKNNgraph( int k, const PointList& points, BoostGraph& EKNNgraph );
void drawBoostGraph( const BoostGraph& graph, const PointList& points );
void drawBoostGraph( const BoostGraph& graph, const PointList& points, const PointList& normals );
void drawWeightedBoostGraph( const BoostGraph& graph, const PointList& points, const Deform::DeformAll_Utility::SparseWeightList& weightList, const std::vector<RealColor>& colors, int showComp );
void drawWeightedBoostGraph( const BoostGraph& graph, const PointList& points, const PointList& normals, const Deform::DeformAll_Utility::SparseWeightList& weightList, const std::vector<RealColor>& colors, int showComp );

#endif
