#include "Colormap.h"
#include "util.h"

using namespace std;

Colormap::Colormap()
{
	makeDefault();
}

void Colormap::makeDefault() {
	//// Default colormap is "autumn" from Matlab
	//CtrlPts.resize(2);
	//CtrlColor.resize(2);
	//CtrlPts[0] = 0; CtrlColor[0] = RealColor(1,0,0);
	//CtrlPts[1] = 1; CtrlColor[1] = RealColor(1,1,0);

	// Default colormap is "jet" from Matlab
	CtrlPts.resize(6);
	CtrlColor.resize(6);
	CtrlPts[0] = 0;     CtrlColor[0] = RealColor(0,0,0.5625);
	CtrlPts[1] = 0.125; CtrlColor[1] = RealColor(0,0,1);
	CtrlPts[2] = 0.375; CtrlColor[2] = RealColor(0,1,1);
	CtrlPts[3] = 0.625; CtrlColor[3] = RealColor(1,1,0);
	CtrlPts[4] = 0.875; CtrlColor[4] = RealColor(1,0,0);
	CtrlPts[5] = 1;     CtrlColor[5] = RealColor(0.5,0.0);
}

bool Colormap::save( const char *str )
{
	// Save the information to a file
	QFile file(str);
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		cout << "Warning in Colormap::save(): Cannot open save file " << str << "!" << endl;
		return false;
	}

	cout << "Saving colormap...";

	QTextStream out(&file);
	for (unsigned int i = 0; i < CtrlPts.size(); i++) {
		out << CtrlPts[i] << ", " << CtrlColor[i] << "\n";
	}
	file.close();

	cout << "done!" << endl;
	return true;
}

bool Colormap::load( const char *str )
{
	QFile file(str);
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		cout << "Warning in Colormap::load(): Cannot open file " << str << "!" << endl;
		return false;
	}

	CtrlPts.clear();
	CtrlColor.clear();
	cout << "Loading colormap...";

	real_type pt, r, g, b;
	QTextStream in(&file);
	while (!in.atEnd()) {
		QString line = in.readLine();
		sscanf_s(qPrintable(line), "%f, [%f,%f,%f]", &pt, &r, &g, &b);
		
		if ((CtrlPts.size() > 0) && (pt <= CtrlPts[CtrlPts.size()-1])) {
			cerr << endl << "Error in Colormap::load(): Control points must be in strictly increasing order!  Creating default colormap." << endl;
			makeDefault();
			return false;
		}
		CtrlPts.push_back(pt);
		CtrlColor.push_back(RealColor(r,g,b));
	}

	file.close();

	cout << "done!" << endl;
	return true;
}

RealColor Colormap::getColor( real_type x ) const
{
	// Clamp the value between 0 and 1
	x = std::max((real_type)0.0, std::min((real_type)1.0, x));

	// Search for the interval
	index_type i = 0;
	while (CtrlPts[i] < x) {
		i++;
	}

	if (i == 0) {
		return CtrlColor[0];
	}
	else if (i > CtrlPts.size()-1) {
		return CtrlColor[CtrlColor.size()-1];
	}
	else {
		// Linearly interpolate
		real_type a, b, len;
		a = CtrlPts[i-1];
		b = CtrlPts[i];
		len = b - a;
		
		return (CtrlColor[i]*((x-a)/(b-a)) + CtrlColor[i-1]*((b-x)/(b-a)));
	}
}
