#ifndef __COLORMAP_H__
#define __COLORMAP_H__

#include "types.h"

class Colormap {
public:
	Colormap();

	bool save( const char *str );
	bool load( const char *str );

	virtual RealColor getColor( real_type x ) const;

private:
	void makeDefault();

	std::vector<real_type> CtrlPts;
	std::vector<RealColor> CtrlColor;
};

class ColormapRange {
public:
	// Constructors
	ColormapRange() : minRange(0.0), invRangeLength(1.0) {
		map = new Colormap();
	}
	ColormapRange( real_type minRange, real_type maxRange ) {
		map = new Colormap();
		updateRange(minRange, maxRange);
	}
	ColormapRange( Colormap& map, real_type minRange, real_type maxRange ) {
		this->map = &map;
		updateRange(minRange, maxRange);
	}

	// Get the color of a value
	virtual RealColor getColor( real_type x ) const {
		return map->getColor((x - minRange) * invRangeLength);
	}

	void updateRange( real_type minRange, real_type maxRange ) {
		this->minRange = minRange;
		invRangeLength = 1.0 / (maxRange - minRange);
	}

	// The colormap itself
	Colormap *map;

private:
	real_type minRange, invRangeLength;
};

#endif
