#include "FileSequence.h"
#include <QRegExp>
#include <QDir>
#include <iostream>
#include <algorithm>

using namespace std;

void splitPath( const QString &path, QString &dir, QString &file )
{
	QRegExp rx("[/\\\\]([^/\\\\]+)$");		/**< Grab the last separator. */
	int pos = 0;

	// Clear the output
	dir.clear(); file.clear();

	if ((pos = rx.indexIn(path, 0)) != -1) {
		dir = path.left(pos + 1);
		file = path.right(path.length() - pos - 1);
	}
}


QString grabFile( const QString &path )
{
	QString dir, file;
	splitPath(path, dir, file);
	return file;
}

QString grabPath( const QString &path )
{
	QString dir, file;
	splitPath(path, dir, file);
	return dir;
}

QString removeExt( const QString &filename )
{
	QString fileMatch = filename;
	QRegExp rx("[.]([^./\\\\]+)$");
	int pos = 0;
	if ((pos = rx.indexIn(filename, 0)) != -1) {
		fileMatch = filename.left(pos);
	}
	return fileMatch;
}


FileSequence::FileSequence() :
startFrame(0), numDigits(0), endFrame(0), skipFrame(1)
{

}

FileSequence::FileSequence( QString filename ) :
startFrame(0), numDigits(0), endFrame(0), skipFrame(1)
{
	construct(filename);
}

FileSequence::FileSequence( const char *filename ) :
startFrame(0), numDigits(0), endFrame(0), skipFrame(1)
{
	construct(QString(filename));
}

FileSequence::FileSequence( std::string filename ) :
startFrame(0), numDigits(0), endFrame(0), skipFrame(1)
{
	construct(QString(filename.c_str()));
}


QString FileSequence::operator[]( index_type i ) const
{
	QString out = format;

	if (hasNumber()) {
		out = out.arg(i, numDigits, 10, QLatin1Char('0'));
	}

	return out;
}

QString FileSequence::operator[]( QString arg ) const
{
	QString out = format;

	if (hasNumber()) {
		out = out.arg(arg);
	}

	return out;
}

bool FileSequence::hasNumber() const
{
	return (numDigits > 0);
}


void FileSequence::construct( QString filename )
{
	//////////////////////////////////////////////////////////////////////////
	// Split the directory and the filename
	QString dir, file;
	splitPath(filename, dir, file);

	if (dir.isEmpty() && file.isEmpty()) {
		return;
	}

	//cout << "Directory: " << qPrintable(dir) << endl;
	//cout << "File: " << qPrintable(file) << endl;

	//////////////////////////////////////////////////////////////////////////
	// Setup replacing the last digit in the filename
	QRegExp rx2("(\\d+)");
	QString match;
	int newPos, pos = 0;

	while ((newPos = rx2.indexIn(file, pos+match.length())) != -1) {
		match = rx2.cap(1);
		pos = newPos;
	}

	if (!match.isEmpty()) {
		startFrame = match.toInt();
		numDigits = match.length();
		format = dir + file.left(pos) + QString("%1") + file.right(file.length() - pos - numDigits);

		//cout << "Match at " << numPos << ": " << qPrintable(match) << endl;
		//cout << "Format: " << qPrintable(format) << endl;
	}
	else {
		// there was no match
		startFrame = 0;
		format = filename;
	}
}

bool FileSequence::valid() const
{
	return !format.isEmpty();
}

idxList FileSequence::autoDetect() const
{
	QString dir, file;
	splitPath(format, dir, file);

	// Get the directory
	if (dir.isEmpty() && file.isEmpty()) {
		return idxList();
	}

	//cout << "Format is: " << qPrintable(format) << endl;
	//cout << "Number of digits is: " << numDigits << endl;
	//cout << "Directory: " << qPrintable(dir) << endl;
	//cout << "File: " << qPrintable(file) << endl;

	// Get a list of all files in the directory
	QDir realDir(dir);
	QStringList dirFiles = realDir.entryList();

	// Find all files that match the format
	idxList result;
	QStringList::const_iterator constIterator;
	for (constIterator = dirFiles.constBegin(); constIterator != dirFiles.constEnd(); ++constIterator) {
		QRegExp rx(file.arg("(\\d+)"));
		//cout << "File: " << qPrintable(*constIterator);

		if (rx.exactMatch(*constIterator)) {
			int fn = rx.cap(1).toInt();
			QString reproduced = file.arg(fn, numDigits, 10, QLatin1Char('0'));

			// Test if we can reproduce the exact string
			if (constIterator->compare(reproduced) == 0) {
				//cout << " -- Matched seq " << fn;
				result.push_back(fn);
			}
		}
		//cout << endl;
	}

	if (result.size()) {
		// Sort the frames
		sort(result.begin(), result.end());

		cout << "*** Detected " << result.size() << " files (" << qPrintable(file) << ")" << endl;
	}

	return result;
}
