#include "CorrespViewer.h"
#include "scene.h"
#include "viewer.h"
#include <algorithm>
#include "FileSequence.h"
#include "../Util/util.h"

using namespace std;

CorrespViewerElement::CorrespViewerElement(CorrespViewer *v, Scene *s, int displayMode) :
parent(v), scene(s), displayMode(displayMode), pointSelect(3), selectionMode_(NONE),
contextMenu(NULL), saveCorrespondencesAct(NULL), loadCorrespondencesAct(NULL),
clearCorrespondencesAct(NULL), showMotionSegmentationAct(NULL), showModelsAct(NULL),
switchCorrespViewTypeAct(NULL), dummyTitleAct(NULL), loadCorrespResultAct(NULL),
saveCorrespResultAct(NULL), showVerifiedRegionsAct(NULL), saveScreenshotAct(NULL),
showExtraEdgesAct(NULL), displaySubsetNum(10), displaySubsetIndex(0), displaySubsetTotal(0),
displaySubsetMode(TRANSFORM_SET_PAIRS), showMotionSegmentationAsPointsAct(NULL),
switchSelectModeAct(NULL), cList(NULL), displayDeformEMResultAct(NULL), displayDeformAllResultAct(NULL), storedFrameIndex(0),
showOtherModelAct(NULL)
{
}

CorrespViewerElement::~CorrespViewerElement()
{
	if (contextMenu) delete contextMenu;
	if (dummyTitleAct) delete dummyTitleAct;
	if (saveCorrespondencesAct) delete saveCorrespondencesAct;
	if (loadCorrespondencesAct) delete loadCorrespondencesAct;
	if (clearCorrespondencesAct) delete clearCorrespondencesAct;
	if (showMotionSegmentationAct) delete showMotionSegmentationAct;
	if (showModelsAct) delete showModelsAct;
	if (switchCorrespViewTypeAct) delete switchCorrespViewTypeAct;
	if (saveCorrespResultAct) delete saveCorrespResultAct;
	if (loadCorrespResultAct) delete loadCorrespResultAct;
	if (showVerifiedRegionsAct) delete showVerifiedRegionsAct;
	if (saveScreenshotAct) delete saveScreenshotAct;
	if (showExtraEdgesAct) delete showExtraEdgesAct;
	if (showMotionSegmentationAsPointsAct) delete showMotionSegmentationAsPointsAct;
	if (switchSelectModeAct) delete switchSelectModeAct;
	if (displayDeformEMResultAct) delete displayDeformEMResultAct;
	if (displayDeformAllResultAct) delete displayDeformAllResultAct;
	if (saveCameraAct) delete saveCameraAct;
	if (loadCameraAct) delete loadCameraAct;
	if (regenerateColorsAct) delete regenerateColorsAct;
	if (showOtherModelAct) delete showOtherModelAct;
}

void CorrespViewerElement::init()
{
	cout << " --> Initializing ";
	switch(displayMode)
	{
	case DISPLAY_SOURCE:
		cout << "Source";
		break;
	case DISPLAY_TARGET:
		cout << "Target";
	    break;
	default:
	    break;
	}
	cout << " Viewer...";

	createContextMenu();

	//////////////////////////////////////////////////////////////////////////
	cout << " view";
	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);

	if (displayMode == DISPLAY_SOURCE) {
		restoreStateFromFile();
	}

	//// Center the view on the source and target
	//real3_type box_min, box_max;
	//for (int d = 0; d < 3; d++) {
	//	box_min[d] = min(scene->source->box_min[d], scene->target->box_min[d]);
	//	box_max[d] = max(scene->source->box_max[d], scene->target->box_max[d]);
	//}
	//setSceneBoundingBox(qglviewer::Vec(box_min[0], box_min[1], box_min[2]),
	//					qglviewer::Vec(box_max[0], box_max[1], box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	//showEntireScene();

	//////////////////////////////////////////////////////////////////////////
	cout << " gl";
	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the max select buffer size
	setSelectBufferSize(100000);

	// Start animation
	startAnimation();

	//////////////////////////////////////////////////////////////////////////
	cout << "...done!" << endl;
}

void CorrespViewerElement::drawSubset()
{
	const ShapeModel *shape = NULL, *sample = NULL;
	idxList *selectionList = NULL;
	FloatArray materialColor;
	if (displayMode == DISPLAY_SOURCE) {
		materialColor = scene->sourceColor;
		selectionList = &parent->selected;
		if (scene->matchCandidates) {
			shape = &scene->matchCandidates->sourceModel;
			sample = scene->matchCandidates->sourceSample;
		}
	}
	else if (displayMode == DISPLAY_TARGET) {
		materialColor = scene->targetColor;
		selectionList = &parent->targetSelected;
		if (scene->matchCandidates) {
			shape = &scene->matchCandidates->targetModel;
			sample = scene->matchCandidates->targetSample;
		}
	}

	if (!shape || !selectionList) {
		return;
	}

	// Draw the entire shape
	setOpenGLMaterialColor(materialColor);
	shape->draw();

	if ((parent->selectMode == SELECT_VERTICES) && selectionList->size()) {
		QString message = tr("Selected %1 vertices").arg(selectionList->size());

		//////////////////////////////////////////////////////////////////////////
		// Draw just the subset
		shape->drawSubset(*selectionList);

		//////////////////////////////////////////////////////////////////////////
		// Draw sample selection
		if (sample) {
			std::vector<bool> vertexMark = scene->matchCandidates->closestSampleIndices(*shape, *sample, *selectionList);

			//glDisable(GL_LIGHTING);
			//glPointSize(5.f);
			//glColor3f(0.5,0.5,0.5);
			//glBegin(GL_POINTS);
			//int numMarked = 0;
			//for(index_type i = 0; i < sample->numVertices(); i++) {
			//	if (vertexMark[i]) {
			//		real3_type v = sample->vertex[i];
			//		glVertex3f(v[0],v[1],v[2]);
			//		numMarked++;
			//	}
			//}
			//glEnd();

			int numMarked = 0;
			setOpenGLMaterialColor(scene->pointColor);
			for(index_type i = 0; i < sample->numVertices(); i++) {
				if (vertexMark[i]) {
					drawSphere(sample->vertex[i], 0.75*scene->pointSize());
					numMarked++;
				}
			}

			message += tr(" (%2 samples)").arg(numMarked);
		}

		drawTextMessage(1, foregroundColor(), message);
	}
}

void CorrespViewerElement::drawTransformedSubset( int transformSet )
{
	bool resultExists = false;

	//////////////////////////////////////////////////////////////////////////
	// Check to see if the required result exists
	if (transformSet == TRANSFORM_SET_PAIRS) {
		resultExists = (scene->matchCandidates && scene->matchCandidates->size());
	}
	else if (transformSet == TRANSFORM_SET_CLUSTERS) {
		resultExists = (scene->matchCandidates && scene->matchCandidates->doneClustering());
	}
	else if (transformSet == TRANSFORM_SET_REGIONS) {
		resultExists = (scene->matchCandidates && scene->matchCandidates->doneVerify());
	}
	if (!resultExists) {
		// If the required result doesn't exist, just draw the untransformed subset
		setOpenGLMaterialColor(scene->sourceColor);
		scene->matchCandidates->sourceModel.drawSubset(parent->selected);
		return;
	}


	// Obtain the set of transformations associated with the selected vertices
	idxList foundSet;

	if (transformSet == TRANSFORM_SET_PAIRS) {
		foundSet = scene->matchCandidates->selectPairsHavingVertices(parent->selected, parent->targetSelected);
	}
	else if (transformSet == TRANSFORM_SET_CLUSTERS) {
		foundSet = scene->matchCandidates->selectClustersHavingVertices(parent->selected, parent->targetSelected);
	}
	else if (transformSet == TRANSFORM_SET_REGIONS) {
		foundSet = scene->matchCandidates->selectRegionsHavingVertices(parent->selected);
	}

	displaySubsetTotal = foundSet.size();

	//////////////////////////////////////////////////////////////////////////
	// Draw the transformed subsets
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glCullFace(GL_BACK);
	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	setOpenGLMaterialColor(scene->sourceColor, 0.7f);

	int startIndex = max(0, min(displaySubsetIndex, displaySubsetTotal-1));
	int endIndex = max(0, min(displaySubsetIndex+displaySubsetNum, displaySubsetTotal));

	for (index_type i = startIndex; i < endIndex; i++) {
		SE3 tf;
		if (transformSet == TRANSFORM_SET_PAIRS) {
			tf = scene->matchCandidates->getPairTransform(foundSet[i]);
		}
		else if (transformSet == TRANSFORM_SET_CLUSTERS) {
			tf = scene->matchCandidates->getClusterTransform(foundSet[i]);
		}
		else if (transformSet == TRANSFORM_SET_REGIONS) {
			tf = scene->matchCandidates->sortedTFList[foundSet[i]];
		}

		glPushMatrix();
		tf.applyGL();
		scene->matchCandidates->sourceModel.drawSubset(parent->selected);
		glPopMatrix();
	}

	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_CULL_FACE);

	//////////////////////////////////////////////////////////////////////////
	// Show text status
	QString message = tr("Showing %4 transformations %1-%2 (subset %3, total %5)").arg(startIndex).arg(endIndex-1).arg(foundSet.size());
	if (transformSet == TRANSFORM_SET_PAIRS) {
		message = message.arg("pair").arg(scene->matchCandidates->size());
	}
	else if (transformSet == TRANSFORM_SET_CLUSTERS) {
		message = message.arg("cluster").arg(scene->matchCandidates->clusterModes.size());
	}
	else if (transformSet == TRANSFORM_SET_REGIONS) {
		message = message.arg("region").arg(scene->matchCandidates->sortedMatchList.size());
	}
	drawTextMessage(2, foregroundColor(), message);
}

void CorrespViewerElement::draw()
{
	//if (!scene->getAnimation()) {
	//	return;
	//}

	if (displayDeformEMResultAct->isChecked()) {
		Deform::DrawOptions drawOptions;

		drawOptions.showSourceModel = false;
		drawOptions.showTargetModel = (displayMode == DISPLAY_SOURCE);
		drawOptions.showDeformedModel = (displayMode == DISPLAY_SOURCE);
		drawOptions.showEffRad = false;
		drawOptions.renderTransparent = false;
		drawOptions.showComponent = -1;
		drawOptions.showGraph = (displayMode == DISPLAY_TARGET);
		drawOptions.showGraphDeformed = true;
		drawOptions.showGraphFaces = true;
		drawOptions.showExtendedGrid = -1;
		drawOptions.showWireframe = false;
		drawOptions.openGLPointSize = 5.0;
		drawOptions.sourceModelColor = RealColor(1.0f, 0.7f, 0.7f);
		drawOptions.deformedModelColor = RealColor(0.7f, 1.0f, 0.7f);
		drawOptions.targetModelColor = RealColor(0.7f, 0.7f, 1.0f);
		drawOptions.ignoreDisplayInterface = (displayMode == DISPLAY_TARGET);

		scene->drawDeformation(drawOptions);

		if (displayMode == DISPLAY_SOURCE) {
			QFont useFont;
			useFont.setBold(true);
			useFont.setPixelSize(20);

			// Draw solver status on bottom of this window
			glColor3f(0.0,0.0,0.0);
			drawText(20, height() - 20, scene->getSolverStatus(), useFont);
			//cout << scene->getSolverStatus().c_str();
		}
	}

	if (displayDeformAllResultAct->isChecked()) {
		Deform::DrawOptions drawOptions = parent->getMasterViewer()->getDeformDrawOptions();
		QString status = scene->getSolverStatus();

		// Change draw state based on status
		if (status.startsWith(tr("Initializing frame"))) {
			// Get frame number
			QString frameNumStr = status.right(status.length() - 19);
			storedFrameIndex = frameNumStr.toInt();			

			//cout << "Changing to frame (" << qPrintable(frameNumStr) << ") = " << storedFrameIndex << endl;
		}

		// Switch to the automatically determined frame
		drawOptions.displayFrame = storedFrameIndex;

		if (displayMode == DISPLAY_TARGET) {
			// Populate option values from the viewer and the scene
			drawOptions.showSourceModel = false;
			drawOptions.showTargetModel = false;
			drawOptions.showDeformedModel = false;
			drawOptions.showSourceCorresp = false;
			drawOptions.showTargetCorresp = false;
			drawOptions.showDeformedCorresp = false;
			drawOptions.showGraph = false;
			drawOptions.showGraphFaces = true;
			drawOptions.showGraphDeformed = true;
			drawOptions.showAllFrames = false;

			if (status.startsWith(tr("Global")) || status.startsWith(tr("Update Reference"))) {
				// Deform grid
				drawOptions.showGraphDeformed = false;
			}
			else {
				drawOptions.showGraphDeformed = true;
			}
		}
		else if (displayMode == DISPLAY_SOURCE) {
			--drawOptions.displayFrame;

			if (storedFrameIndex > 0) {
				drawOptions.showSourceModel = false;
				drawOptions.showTargetModel = true;
				drawOptions.showDeformedModel = true;
			}
			else {
				drawOptions.showSourceModel = true;
				drawOptions.showTargetModel = false;
				drawOptions.showDeformedModel = false;
			}

			if (status.startsWith(tr("Global")) || status.startsWith(tr("Update Reference"))) {
				drawOptions.showAllFrames = true;
				drawOptions.showGraphDeformed = false;
			}
			else {
				drawOptions.showAllFrames = false;
				drawOptions.showGraphDeformed = true;
			}

			drawOptions.showJoints = true;
			drawOptions.showGraph = false;
		}

		scene->drawDeformation(drawOptions);

		if (displayMode == DISPLAY_SOURCE) {
			QFont useFont;
			useFont.setBold(true);
			useFont.setPixelSize(20);

			// Draw solver status on bottom of this window
			glColor3f(0.0,0.0,0.0);
			drawText(20, height() - 20, status, useFont);
			//cout << scene->getSolverStatus().c_str();
		}
	}

	if (showModelsAct->isChecked()) {
		glEnable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);

		if (displayMode == DISPLAY_SOURCE) {

			drawSubset();

			if (parent->selectMode == SELECT_CORRESP) {
				// Display source correspondences
				if (cList && selectionMode_ != MOVE) {
					
					setOpenGLMaterialColor(scene->selectedColor);
					PointList_Utility::drawWithWeight(cList->sourcePoints, POINT_SPHERES, parent->selected, cList->weight, scene->pointSize());
					setOpenGLMaterialColor(scene->pointColor);
					PointList_Utility::drawWithWeight(cList->sourcePoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
			}

			if (showOtherModelAct->isChecked()) {
				// Show target model
				setOpenGLMaterialColor(scene->targetColor);
				if (scene->matchCandidates) {
					scene->matchCandidates->targetModel.draw();
				}
			}
		}
		else if (displayMode == DISPLAY_TARGET) {

			drawSubset();

			if ((parent->selectMode == SELECT_VERTICES) && parent->selected.size()) {
				drawTransformedSubset(displaySubsetMode);
			}

			if (parent->selectMode == SELECT_CORRESP) {
				// Display target correspondences
				if (cList && selectionMode_ != MOVE) {
					setOpenGLMaterialColor(scene->selectedColor);
					PointList_Utility::drawWithWeight(cList->targetPoints, POINT_SPHERES, parent->selected, cList->weight, scene->pointSize());
					setOpenGLMaterialColor(scene->pointColor);
					PointList_Utility::drawWithWeight(cList->targetPoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
			}					
		}
	}

	if (showMotionSegmentationAct->isChecked()) {
		showMotionSegmentation();
	}
	if (showVerifiedRegionsAct->isChecked()) {
		showVerifiedRegions();
	}
	if (showClusterMembersAct->isChecked()) {
		showClusterMembers();
	}
	if (showExtraEdgesAct->isChecked()) {
		drawExtraEdges();
	}


	// Draws rectangular selection area. Could be done in postDraw() instead.
	if (selectionMode_ == TOGGLE)
		drawSelectionRectangle();
	else if (selectionMode_ == ADD) {
		setOpenGLMaterialColor(scene->selectedColor);
		
		if (displayMode == DISPLAY_SOURCE && parent->setNewSourcePt) {
			drawSphere(parent->newSourcePt, scene->pointSize());
		}
		else if (displayMode == DISPLAY_TARGET && parent->setNewTargetPt) {
			drawSphere(parent->newTargetPt, scene->pointSize());
		}
	}
}

void CorrespViewerElement::updateDisplay()
{
	updateGL();
}

void CorrespViewerElement::drawWithNames()
{
	if (scene->matchCandidates) {
		cout << "Drawing with names, mode = " << parent->selectMode << "..." << endl;
		if (parent->selectMode == SELECT_VERTICES) {
			// Draw mesh vertices
			glPointSize(5.0);
			if (displayMode == DISPLAY_SOURCE) {
				cout << "Drawing source model with names..." << endl;
				PointList_Utility::drawWithNames(scene->matchCandidates->sourceModel.vertex, POINT_SQUARES);
			}
			else if (displayMode == DISPLAY_TARGET) {
				cout << "Drawing target model with names..." << endl;
				PointList_Utility::drawWithNames(scene->matchCandidates->targetModel.vertex, POINT_SQUARES);
			}
		}
		else if (parent->selectMode == SELECT_CORRESP) {
			// Draw correspondences
			if (cList) {
				if (displayMode == DISPLAY_SOURCE) {
					PointList_Utility::drawWithNamesWithWeight(cList->sourcePoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
				else if (displayMode == DISPLAY_TARGET) {
					PointList_Utility::drawWithNamesWithWeight(cList->targetPoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
			}
		}
	}
}

void CorrespViewerElement::drawSelectionRectangle() const
{
	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_BLEND);

	glColor4f(0.0f, 0.0f, 0.3f, 0.3f);
	glBegin(GL_QUADS);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glLineWidth(2.0);
	glColor4f(0.4f, 0.4f, 0.5f, 0.5f);
	glBegin(GL_LINE_LOOP);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();
}

void CorrespViewerElement::mousePressEvent( QMouseEvent *e )
{
	// Start selection. Mode is ADD with Shift key and TOGGLE with Alt key.
	rectangle_ = QRect(e->pos(), e->pos());

	// Reset the add selection mode if an add doesn't immediately follow with another add.
	if (selectionMode_ == ADD) {
		if (((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier))
			|| ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ShiftModifier))) {
			parent->setNewSourcePt = parent->setNewTargetPt = false;
			selectionMode_ = NONE;
		}
	}

	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier)) {
		selectionMode_ = TOGGLE;
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ShiftModifier)) {
		// Define selection window dimensions
		setSelectRegionWidth(1);
		setSelectRegionHeight(1);

		// Compute rectangle center and perform selection
		parent->selected.clear();
		select(rectangle_.center());

		if (parent->selected.size() == 1) {
			selectionMode_ = MOVE;
		}
		else {
			selectionMode_ = NONE;
		}
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		selectionMode_ = ADD;
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void CorrespViewerElement::mouseMoveEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Updates rectangle_ coordinates and redraws rectangle
		rectangle_.setBottomRight(e->pos());
		parent->updateDisplay();
	}
	else if (selectionMode_ == MOVE) {
		updateGL();

		// Move the selected vertex along the shape
		bool found;
		Geometry::Vec3f pickedPoint = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));

		if (cList && found) {
			cout << "Setting position " << parent->selected[0] << " to " << pickedPoint << endl;

			// Modify the selected point's position
			if (displayMode == DISPLAY_SOURCE) {
				cList->setSourcePos(parent->selected[0], pickedPoint);
			}
			else if (displayMode == DISPLAY_TARGET) {
				cList->setTargetPos(parent->selected[0], pickedPoint);
			}
		}

		parent->updateDisplay();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void CorrespViewerElement::mouseReleaseEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Actual selection on the rectangular area.
		// Possibly swap left/right and top/bottom to make rectangle_ valid.
		rectangle_ = rectangle_.normalized();

		int rw, rh;
		rw = rectangle_.width();
		rh = rectangle_.height();

		if (rw <= 0) rw = 1;
		if (rh <= 0) rh = 1;

		// Define selection window dimensions
		setSelectRegionWidth(rw);
		setSelectRegionHeight(rh);

		// Compute rectangle center and perform selection
		select(rectangle_.center());

		// reset selection mode
		selectionMode_ = NONE;

		// Update display to show new selected objects
		parent->updateDisplay();
	}
	else if (selectionMode_ == MOVE) {
		// reset selection mode
		selectionMode_ = NONE;

		parent->updateDisplay();
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		updateGL();

		// Add the picked point
		bool found;
		Geometry::Vec3f pickedPoint = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));

		if (found) {
			//cout << "Adding position " << pickedPoint << endl;

			// Modify the selected point's position
			if (displayMode == DISPLAY_SOURCE) {
				parent->newSourcePt = pickedPoint;
				parent->setNewSourcePt = true;
			}
			else if (displayMode == DISPLAY_TARGET) {
				parent->newTargetPt = pickedPoint;
				parent->setNewTargetPt = true;
			}
		}

		// If both the new source and target point is set, then add the new correspondence
		if (parent->setNewSourcePt && parent->setNewTargetPt) {
			cList->add(parent->newSourcePt, parent->newTargetPt);
			parent->setNewSourcePt = parent->setNewTargetPt = false;
			selectionMode_ = NONE;
			cout << "Adding point pair..." << endl;
		}

		parent->updateDisplay();
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void CorrespViewerElement::endSelection( const QPoint& )
{
	// Flush GL buffers
	glFlush();

	// Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
	GLint nbHits = glRenderMode(GL_RENDER);

	cout << "Number of objects selected: " << nbHits << endl;

	// Interpret results : each object created 4 values in the selectBuffer().
	// (selectBuffer())[4*i+3] is the id pushed on the stack.
	if (nbHits > 0) {
		// If the rectangle is small, then select a single object of minimum depth.
		if ((rectangle_.width() < pointSelect) || (rectangle_.height() < pointSelect)) {
			GLuint zMin = (selectBuffer())[1];
			int minId = (selectBuffer())[3];

			for (int i=1; i<nbHits; ++i) {
				if ((selectBuffer())[4*i+1] < zMin) {
					zMin = (selectBuffer())[4*i+1];
					minId = (selectBuffer())[4*i+3];
				}
			}

			if (displayMode == DISPLAY_SOURCE) {
				parent->toggleIdSelection(minId);
			}
			else if (displayMode == DISPLAY_TARGET) {
				parent->toggleTargetIdSelection(minId);
			}
		}
		else {
			for (int i=0; i<nbHits; ++i) {
				if (displayMode == DISPLAY_SOURCE) {
					parent->toggleIdSelection((selectBuffer())[4*i+3]);
				}
				else if (displayMode == DISPLAY_TARGET) {
					parent->toggleTargetIdSelection((selectBuffer())[4*i+3]);
				}
			}
		}
	}
	else {
		// Clear the selection
		if (displayMode == DISPLAY_SOURCE) {
			parent->selected.clear();
		}
		else if (displayMode == DISPLAY_TARGET) {
			parent->targetSelected.clear();
		}
	}
}

void CorrespViewerElement::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_Delete) && (modifiers==Qt::NoButton)) {
		// Remove the selected points from the correspondence
		if (cList) {
			cList->remove(parent->selected);
			parent->selected.clear();
			parent->updateDisplay();
		}
	}
	else if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = parent->geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Backspace) && (modifiers==Qt::NoButton)) {
		switchCorrespViewType();
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_P) && (modifiers==Qt::NoButton)) {
		parent->updateDisplay();
		saveScreenshot(false);
	}
	else if ((e->key()==Qt::Key_0) && (modifiers==Qt::KeypadModifier)) {
		// Reset the display subset index
		displaySubsetIndex = 0;
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier)) {
		// Increment the current display subset
		displaySubsetIndex += displaySubsetNum;
		if (displaySubsetIndex >= displaySubsetTotal) {
			displaySubsetIndex = 0;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier)) {
		// Decrement the current display subset
		displaySubsetIndex -= displaySubsetNum;
		if (displaySubsetIndex < 0) {
			displaySubsetIndex = displaySubsetTotal - displaySubsetNum;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Asterisk) && (modifiers==Qt::KeypadModifier)) {
		// Increment the current display number
		displaySubsetNum++;
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Slash) && (modifiers==Qt::KeypadModifier)) {
		// Decrement the current display number
		displaySubsetNum--;
		if (displaySubsetNum < 1) {
			displaySubsetNum = 1;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::NoButton)) {
		displaySubsetMode++;
		if (displaySubsetMode > 3) {
			displaySubsetMode = 3;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::NoButton)) {
		displaySubsetMode--;
		if (displaySubsetMode < 1) {
			displaySubsetMode = 1;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::CTRL)) {
		scene->regenerateColors();
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::ALT)) {
		resetView();
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}

void CorrespViewerElement::createContextMenu()
{
	// Create menu bar.
	QString menuTitle;

	if (displayMode == DISPLAY_SOURCE) {
		menuTitle = tr("&Source Model View");
	}
	else if (displayMode == DISPLAY_TARGET) {
		menuTitle = tr("&Target Model View");
	}

	contextMenu = new QMenu(tr("&Menu"), this);

	dummyTitleAct = new QAction(menuTitle, this);
	contextMenu->addAction(dummyTitleAct);
	dummyTitleAct->setDisabled(true);

	contextMenu->addSeparator();

	saveCorrespondencesAct = new QAction(tr("&Save Correspondences..."), this);
	contextMenu->addAction(saveCorrespondencesAct);
	connect(saveCorrespondencesAct, SIGNAL(triggered()), this, SLOT(saveCorrespondences()));

	loadCorrespondencesAct = new QAction(tr("&Load Correspondences..."), this);
	contextMenu->addAction(loadCorrespondencesAct);
	connect(loadCorrespondencesAct, SIGNAL(triggered()), this, SLOT(loadCorrespondences()));

	clearCorrespondencesAct = new QAction(tr("&Clear Correspondences..."), this);
	contextMenu->addAction(clearCorrespondencesAct);
	connect(clearCorrespondencesAct, SIGNAL(triggered()), this, SLOT(clearCorrespondences()));

	contextMenu->addSeparator();

	showModelsAct = new QAction(tr("&Show Models..."), this);
	showModelsAct->setCheckable(true);
	contextMenu->addAction(showModelsAct);
	showModelsAct->setChecked(true);

	showClusterMembersAct = new QAction(tr("&Show Cluster Members..."), this);
	showClusterMembersAct->setCheckable(true);
	contextMenu->addAction(showClusterMembersAct);

	showVerifiedRegionsAct = new QAction(tr("&Show Verified Regions..."), this);
	showVerifiedRegionsAct->setCheckable(true);
	contextMenu->addAction(showVerifiedRegionsAct);

	showMotionSegmentationAct = new QAction(tr("&Show Motion Clustering Result..."), this);
	showMotionSegmentationAct->setCheckable(true);
	contextMenu->addAction(showMotionSegmentationAct);

	switchCorrespViewTypeAct = new QAction(tr("&Switch Correspondence Direction..."), this);
	contextMenu->addAction(switchCorrespViewTypeAct);
	connect(switchCorrespViewTypeAct, SIGNAL(triggered()), this, SLOT(switchCorrespViewType()));

	showExtraEdgesAct = new QAction(tr("&Show Extra Smoothness Constraints..."), this);
	showExtraEdgesAct->setCheckable(true);
	contextMenu->addAction(showExtraEdgesAct);
	showExtraEdgesAct->setChecked(true);

	showMotionSegmentationAsPointsAct = new QAction(tr("&Show Motion Clustering Result As Points..."), this);
	showMotionSegmentationAsPointsAct->setCheckable(true);
	contextMenu->addAction(showMotionSegmentationAsPointsAct);

	switchSelectModeAct = new QAction(tr("&Switch Mouse Selection Mode..."), this);
	contextMenu->addAction(switchSelectModeAct);
	connect(switchSelectModeAct, SIGNAL(triggered()), parent, SLOT(SwitchSelectMode()));

	displayDeformEMResultAct = new QAction(tr("&Display DeformEM Result..."), this);
	displayDeformEMResultAct->setCheckable(true);
	contextMenu->addAction(displayDeformEMResultAct);
	displayDeformEMResultAct->setChecked(false);

	displayDeformAllResultAct = new QAction(tr("&Display DeformAll Result..."), this);
	displayDeformAllResultAct->setCheckable(true);
	contextMenu->addAction(displayDeformAllResultAct);
	displayDeformAllResultAct->setChecked(true);

	showOtherModelAct = new QAction(tr("&Show Other Model"), this);
	showOtherModelAct->setCheckable(true);
	contextMenu->addAction(showOtherModelAct);
	showOtherModelAct->setChecked(false);

	contextMenu->addSeparator();

	saveCorrespResultAct = new QAction(tr("&Save Correspondence Result..."), this);
	contextMenu->addAction(saveCorrespResultAct);
	connect(saveCorrespResultAct, SIGNAL(triggered()), this, SLOT(saveCorrespResult()));

	loadCorrespResultAct = new QAction(tr("&Load Correspondence Result..."), this);
	contextMenu->addAction(loadCorrespResultAct);
	connect(loadCorrespResultAct, SIGNAL(triggered()), this, SLOT(loadCorrespResult()));

	saveScreenshotAct = new QAction(tr("&Save Screenshot (Press P)"), this);
	contextMenu->addAction(saveScreenshotAct);
	connect(saveScreenshotAct, SIGNAL(triggered()), this, SLOT(saveScreenshot()));

	saveCameraAct = new QAction(tr("&Save Camera..."), this);
	contextMenu->addAction(saveCameraAct);
	connect(saveCameraAct, SIGNAL(triggered()), this, SLOT(saveCamera()));

	loadCameraAct = new QAction(tr("&Load Camera..."), this);
	contextMenu->addAction(loadCameraAct);
	connect(loadCameraAct, SIGNAL(triggered()), this, SLOT(loadCamera()));

	regenerateColorsAct = new QAction(tr("&Regenerate Colors..."), this);
	contextMenu->addAction(regenerateColorsAct);
	connect(regenerateColorsAct, SIGNAL(triggered()), this, SLOT(regenerateColors()));
}


void CorrespViewerElement::saveCorrespondences()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the correspondeces...", scene->lastDir, QString("*.txt"));
	if (savefileName.length() == 0) {
		cout << "Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	if (cList) {
		cList->save(std::string(qPrintable(savefileName)));
	}
}

void CorrespViewerElement::loadCorrespondences()
{
	QString loadfileName;
	loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the correspondeces...", scene->lastDir, tr("*.txt"));
	if (loadfileName.length() == 0) {
		cout << "Loading canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(loadfileName);
	}

	if (cList) {
		cList->load(std::string(qPrintable(loadfileName)));
	}
}

void CorrespViewerElement::clearCorrespondences()
{
	if (cList) {
		cList->clear();
		parent->selected.clear();
	}
}

void CorrespViewerElement::switchCorrespViewType()
{
	if (parent->showCorrespType == SHOW_SOURCE_TO_TARGET) {
		parent->showCorrespType = SHOW_TARGET_TO_SOURCE;
	}
	else if (parent->showCorrespType == SHOW_TARGET_TO_SOURCE) {
		parent->showCorrespType = SHOW_SOURCE_TO_TARGET;
	}
}

void CorrespViewerElement::showClusterMembers()
{
	if (!scene->matchCandidates || !scene->matchCandidates->doneClustering()) {
		return;
	}

	// After the mean-shift clustering:
	// For each cluster, show the correspondence pairs belonging to that cluster
	const MatchCandidates &candies = *scene->matchCandidates;
	const idxList &clusterLabels = candies.clusterLabels;
	const PointListN &clusterModes = candies.clusterModes;

	if (scene->clusterColors.size() < clusterModes.size()) {
		scene->generateClusterColors(clusterModes.size());
	}

	// Plot the points with the label colors
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);

	for (index_type i = 0; i < candies.size(); i++) {
		index_type label = candies.clusterLabels[i];

		// Check to see if it passes the minimum cluster size test
		if (!candies.validCluster(label) ||
			!(((scene->showClusterMode == 0) && (candies.clusterSizes[label] > scene->minVerifyClusterSize)) ||
				(scene->showClusterMode == label)))
		{
			continue;
		}


		setOpenGLColor(scene->getClusterColor(label));

		real3_type v;
		if (displayMode == DISPLAY_SOURCE) {
			v = candies.sourceSample->vertex[candies.sourceIndex[i]];
		}
		else if (displayMode == DISPLAY_TARGET) {
			v = candies.targetSample->vertex[candies.targetIndex[i]];
		}
		glVertex3f(v[0],v[1],v[2]);
	}
	glEnd();

	// Count the number of displayed clusters
	int numClustersShown = 0;
	for (index_type i = 0; i < clusterModes.size(); i++) {
		// Check to see if it passes the minimum cluster size test
		if (candies.validCluster(i) && (candies.clusterSizes[i] >= scene->minVerifyClusterSize)) {
			numClustersShown++;
		}
	}
	drawTextMessage(2, foregroundColor(), tr("Showing %1/%2 clusters").arg(numClustersShown).arg(clusterModes.size()));
}

void CorrespViewerElement::showVerifiedRegions()
{
	if (!scene->matchCandidates || !scene->matchCandidates->doneVerify()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Draw each verified region
	const std::vector<CorrespondenceList> &clist = scene->matchCandidates->matchList;
	size_type numRegions = clist.size();

	if (scene->clusterColors.size() < numRegions) {
		// If not enough colors, generate more
		scene->generateClusterColors(numRegions);
	}

	glDisable(GL_LIGHTING);
	glPointSize(3.0);
	glBegin(GL_POINTS);
	int numRegionsShown = 0;
	for (index_type ci = 0; ci < numRegions; ci++) {

		if (((scene->showClusterMode == 0) && (clist[ci].size() > scene->minMatchSize_VR))
			|| (scene->showClusterMode == ci)) {
			setOpenGLColor(scene->getClusterColor(ci));

			const CorrespondenceList &list = clist[ci];
			for (index_type i = 0; i < list.size(); i++) {
				real3_type temp;

				// Draw corresponding target vertex
				if (displayMode == DISPLAY_SOURCE) {
					temp = list.sourcePos(i);
				}
				else if (displayMode == DISPLAY_TARGET) {
					temp = list.targetPos(i);
				}
				glVertex3f(temp[0],temp[1],temp[2]);
			}

			numRegionsShown++;
		}
	}
	glEnd();

	//// Draw closest points
	//glColor3f(1,1,1);
	//glBegin(GL_POINTS);
	//const PointList &closestPoints = scene->matchCandidates->closestPoints;
	//for (index_type i = 0; i < closestPoints.size(); i++) {
	//	glVertex3f(closestPoints[i][0],closestPoints[i][1],closestPoints[i][2]);
	//}
	//glEnd();

	// Display status string
	drawTextMessage(2, foregroundColor(), tr("Showing %1/%2 verified regions").arg(numRegionsShown).arg(numRegions));
}


void CorrespViewerElement::showMotionSegmentation()
{
	if (!scene->clusterResultExists()) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Visualize the graph cut labelling result
	const std::vector<index_type> *p_labels;
	const ShapeModel *p_shape;

	if (parent->showCorrespType == SHOW_SOURCE_TO_TARGET) {
		p_labels = &(scene->matchCandidates->graphCutLabels_source);
		p_shape = &(scene->matchCandidates->sourceModel);
	}
	else if (parent->showCorrespType == SHOW_TARGET_TO_SOURCE) {
		p_labels = &(scene->matchCandidates->graphCutLabels_target);
		p_shape = &(scene->matchCandidates->targetModel);
	}

	size_type numRegions = scene->matchCandidates->sortedTFList.size();
	if (scene->clusterColors.size() < numRegions) {
		// If not enough colors, generate more
		scene->generateClusterColors(numRegions);
	}

	// If it's a mesh, color the entire mesh
	if (!showMotionSegmentationAsPointsAct->isChecked() && (p_shape->shapeType == SHAPE_MESH)) {
		glEnable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		//glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		Mesh3_real &mesh = *((Mesh3_real*)p_shape);

		glBegin(GL_TRIANGLES);
		for(size_type t = 0, t_end = mesh.triangle_vertex_index.size(); t < t_end; t++) {
			const index3_type& tvi = mesh.triangle_vertex_index[t];

			// Check that all three labels not dummy
			bool noDummy = true;
			for(size_type i = 0;i < 3;i++) {
				index_type label = (*p_labels)[tvi[i]];

				if (label >= numRegions) {
					noDummy = false;
				}
			}

			// Draw the triangle
			if (noDummy) {
				for(size_type i = 0;i < 3;i++) {
					index_type label = (*p_labels)[tvi[i]];
					real3_type v = mesh.vertex[tvi[i]];
					real3_type n = mesh.normal[tvi[i]];

					if ((displayMode == DISPLAY_SOURCE) && (parent->showCorrespType == SHOW_TARGET_TO_SOURCE)) {
						// Apply inverse transformation
						v = scene->matchCandidates->sortedInverseTFList[label] * v;
						n = scene->matchCandidates->sortedInverseTFList[label].rot * n;
					}
					else if ((displayMode == DISPLAY_TARGET) && (parent->showCorrespType == SHOW_SOURCE_TO_TARGET)) {
						// Apply forward transformation
						v = scene->matchCandidates->sortedTFList[label] * v;
						n = scene->matchCandidates->sortedTFList[label].rot * n;
					}

					setOpenGLMaterialColor(scene->getClusterColor(label), 0.7f);
					glNormal3f(n[0],n[1],n[2]);
					glVertex3f(v[0],v[1],v[2]);
				}
			}
		}
		glEnd();

		//glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}
	else {
		// Otherwise, just display the points
		glDisable(GL_LIGHTING);
		//glEnable(GL_LIGHTING);
		for (index_type vi = 0; vi < p_shape->numVertices(); vi++) {
			real3_type temp = p_shape->vertex[vi];
			index_type label = (*p_labels)[vi];

			if (label < numRegions) {
				if ((displayMode == DISPLAY_SOURCE) && (parent->showCorrespType == SHOW_TARGET_TO_SOURCE)) {
					// Apply inverse transformation
					temp = scene->matchCandidates->sortedInverseTFList[label] * temp;
				}
				else if ((displayMode == DISPLAY_TARGET) && (parent->showCorrespType == SHOW_SOURCE_TO_TARGET)) {
					// Apply forward transformation
					temp = scene->matchCandidates->sortedTFList[label] * temp;
				}

				setOpenGLColor(scene->getClusterColor(label));
				glPointSize(6.f);
				glBegin(GL_POINTS);
				glVertex3f(temp[0],temp[1],temp[2]);
				glEnd();
				//setOpenGLMaterialColor(scene->getClusterColor(label));
				//drawSphere(temp, 0.5*scene->pointSize());
			}
			else {
				// For dummy label
				if (((displayMode == DISPLAY_SOURCE) && (parent->showCorrespType == SHOW_SOURCE_TO_TARGET)) ||
					((displayMode == DISPLAY_TARGET) && (parent->showCorrespType == SHOW_TARGET_TO_SOURCE))) {
					setOpenGLColor(foregroundColor().rgba());
					glPointSize(10.f);
					glBegin(GL_POINTS);
					glVertex3f(temp[0],temp[1],temp[2]);
					glEnd();
				}
			}
		}
	}

	// Display status string
	QString message("Showing correspondences: ");
	QString sourceName, targetName;
	sourceName = scene->matchCandidates->sourceModel.getFilename();
	targetName = scene->matchCandidates->targetModel.getFilename();

	if (displayMode == DISPLAY_SOURCE) {
		message += sourceName;
		if (parent->showCorrespType == SHOW_TARGET_TO_SOURCE) {
			// We applied inverse transformation
			message += tr(" (transformed from target)");
		}
	}
	else if (displayMode == DISPLAY_TARGET) {
		message += targetName;
		if (parent->showCorrespType == SHOW_SOURCE_TO_TARGET) {
			// We applied forward transformation
			message += tr(" (transformed from source)");
		}
	}

	
	drawTextMessage(2, foregroundColor(), message);
}

void CorrespViewerElement::drawExtraEdges()
{
	if (!scene->matchCandidates) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Draw any extra smoothness constraints
	const ShapeModel *p_shape;
	const EdgeList *p_edgeList;
	if (displayMode == DISPLAY_SOURCE) {
		p_shape = &scene->matchCandidates->sourceModel;
		p_edgeList = &scene->matchCandidates->sourceExtraEdges;
	}
	else if (displayMode == DISPLAY_TARGET) {
		p_shape = &scene->matchCandidates->targetModel;
		p_edgeList = &scene->matchCandidates->targetExtraEdges;
	}

	// Draw the lines
	glDisable(GL_LIGHTING);
	glLineWidth(1.f);
	setOpenGLColor(foregroundColor().rgba());
	glBegin(GL_LINES);
	for (index_type i = 0; i < p_edgeList->size(); i++) {
		real3_type v1, v2;

		v1 = p_shape->vertex[(*p_edgeList)[i].vi.x];
		v2 = p_shape->vertex[(*p_edgeList)[i].vi.y];

		glVertex3f(v1[0], v1[1], v1[2]);
		glVertex3f(v2[0], v2[1], v2[2]);
	}
	glEnd();
}

void CorrespViewerElement::DisplayCorrespondenceResult( bool show )
{
	showMotionSegmentationAct->setChecked(show);
}

void CorrespViewerElement::loadCorrespResult()
{
	if (scene->clusterResultExists()) {
		// Prompt to erase current clustering result
		QMessageBox::StandardButton reply = QMessageBox::question(this, tr("Load correspondence result"),
			tr("Are you sure you want to erase the current correspondence result?"),
			QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);

		if ((reply == QMessageBox::No) || (reply == QMessageBox::Cancel)) {
			return;
		}
	}

	// Otherwise load the saved result
	QString fileName = QFileDialog::getOpenFileName(this, tr("Load Result File"), scene->lastDir, tr("Data (*.data)")); 

	if (fileName.length()) {
		scene->lastDir = grabPath(fileName);
		scene->initMatchCandidates(*scene->source, *scene->target, true);
		scene->matchCandidates->loadGraphCutResult(qPrintable(fileName));

		// Display the result
		parent->DisplayCorrespondenceResult(true);
		parent->updateDisplay();
	}
}

void CorrespViewerElement::saveCorrespResult()
{
	if (scene->clusterResultExists()) {
		// Prompt for an output file name
		QString fileName = QFileDialog::getSaveFileName(this, tr("Save Result File"), scene->lastDir, tr("Data (*.data)")); 

		if (fileName.length()) {
			scene->lastDir = grabPath(fileName);

			// Save the correspondence result
			scene->matchCandidates->saveGraphCutResult(qPrintable(fileName));
		}
	}
}

void CorrespViewerElement::saveScreenshot( bool promptFile /*= true */ )
{
	QString saveFile;

	if (promptFile) {
		saveFile = QFileDialog::getSaveFileName(this, tr("Save Screenshot"), scene->lastDir, tr("png (*.png)")); 
		if (saveFile.length()) {
			scene->lastDir = grabPath(saveFile);
		}
	}
	else {
		// Use the autosave filename (replace extension w/ png)
		saveFile = tr(scene->autoSaveCorrespondenceResult.c_str());
		saveFile.replace(tr(".data"), tr("_shot.png"));
	}

	if (!QFile::exists(saveFile)) {
		// Grab both views and save it in a file
		QPixmap::grabWindow(parent->winId()).toImage().save(saveFile);

		// Show status message
		displayMessage(tr("Took screenshot!"));
	}
}

void CorrespViewerElement::drawTextMessage( int i, QColor color, QString text )
{
	// Draw a text message on the screen
	// Draw it on column i
	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_CULL_FACE);

	glColor4f(color.redF(), color.greenF(), color.blueF(), 1.0);
	renderText(10, height() - 12*i, text);
}

void CorrespViewerElement::animate()
{
	// Nothing for now
}

void CorrespViewerElement::customSaveSnapshot()
{
	cout << "Saving snapshot to file: " << qPrintable(customSaveSnapshotFile) << endl << flush;
	saveSnapshot(customSaveSnapshotFile);
}

void CorrespViewerElement::saveScreenshot2( const QString &filename )
{
	customSaveSnapshotFile = filename;
	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal
	updateGL();
	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}

void CorrespViewerElement::resetView()
{
	// Get scene's bounding box and update
	std::pair<real3_type,real3_type> bbox = scene->getSolverDisplayBBox();
	real3_type box_min = bbox.first;
	real3_type box_max = bbox.second;

	setSceneBoundingBox(qglviewer::Vec(box_min[0], box_min[1], box_min[2]),
		qglviewer::Vec(box_max[0], box_max[1], box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	showEntireScene();

	if (displayMode == DISPLAY_SOURCE) {
		cout << "Source viewer ";
	}
	else if (displayMode == DISPLAY_TARGET) {
		cout << "Target viewer ";
	}
	cout << "showing entire scene at: " << box_min << ", " << box_max << endl;

	parent->updateDisplay();
}

void CorrespViewerElement::loadCamera( const QString& file/* = QString()*/ )
{
	QString filename = file;

	if (filename.length() == 0) {
		// Get the save file name
		filename = QFileDialog::getOpenFileName(this, "Choose a file name to load the display parameters...", QString(), QString("*.xml"));
		if (filename.length() == 0) {
			return;
		}
	}

	// Load DOM from file
	QDomDocument document;
	QFile f(filename);
	if (f.open(QIODevice::ReadOnly)) {
		document.setContent(&f);
		f.close();
	}

	// Parse the DOM tree
	QDomElement main = document.documentElement();
	camera()->initFromDOMElement(main);
}

void CorrespViewerElement::saveCamera()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the camera parameters...", QString(), QString("*.xml"));
	if (savefileName.length() == 0) {
		return;
	}

	QDomDocument document("myCamera");
	document.appendChild( camera()->domElement("Camera", document) );

	QFile f(savefileName);
	if (f.open(QIODevice::WriteOnly))
	{
		QTextStream out(&f);
		document.save(out, 2);
	}
}

void CorrespViewerElement::regenerateColors()
{
	if (scene->matchCandidates) {
		size_type numRegions = scene->matchCandidates->sortedTFList.size();
		scene->generateClusterColors(numRegions);
	}
}

CorrespViewer::CorrespViewer( Scene *s, Viewer *v ) :
sourceView(NULL), targetView(NULL), scene(s), sceneChanged(false), masterViewer(v),
setNewSourcePt(false), setNewTargetPt(false), selectMode(SELECT_VERTICES),
showCorrespType(SHOW_SOURCE_TO_TARGET)
{
	init();
}

CorrespViewer::~CorrespViewer()
{
	// Restore camera
	targetView->setCamera(tempCamera);
	sourceView->close();
	targetView->close();
	delete sourceView;
	delete targetView;

	// Detach from subject
	scene->Detach(this);
}

void CorrespViewer::init()
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	// Set the window title.
	setWindowTitle("Correspondence Viewer");

	cout << "Constructing correspondence viewer..." << endl;

	sourceView = new CorrespViewerElement(this, scene, DISPLAY_SOURCE);
	targetView = new CorrespViewerElement(this, scene, DISPLAY_TARGET);

	connect(this, SIGNAL(updateViews()), sourceView, SLOT(updateDisplay()));
	connect(this, SIGNAL(updateViews()), targetView, SLOT(updateDisplay()));

	// Set the same camera
	tempCamera = targetView->camera();
	targetView->setCamera(sourceView->camera());

	// Attach to subject and master viewer
	scene->Attach(this);

	// Initialize the elements of the splitter
	this->addWidget(sourceView);
	this->addWidget(targetView);

	// Set the window geometry
	setGeometry(10, 120, 1260, 560);
}

void CorrespViewer::Update( Subject* theChangedSubject )
{
	// Implement updating of OpenGL here by setting a variable and looping in an animation loop
	sceneMutex.lock();
	sceneChanged = true;
	sceneMutex.unlock();

	//cerr << "Updated by scene at: " << theChangedSubject << endl;
	emit updateViews();
}

void CorrespViewer::updateDisplay() {
	emit updateViews();
}

void CorrespViewer::toggleIdSelection( index_type id )
{
	idxList::iterator found = std::find(selected.begin(), selected.end(), id);
	if (found == selected.end()) {
		selected.push_back(id);
	}
	else {
		selected.erase(found);
	}
}

void CorrespViewer::toggleTargetIdSelection( index_type id )
{
	idxList::iterator found = std::find(targetSelected.begin(), targetSelected.end(), id);
	if (found == targetSelected.end()) {
		targetSelected.push_back(id);
	}
	else {
		targetSelected.erase(found);
	}
}

void CorrespViewer::DisplayCorrespondenceResult( bool show )
{
	sourceView->DisplayCorrespondenceResult(show);
	targetView->DisplayCorrespondenceResult(show);
}

void CorrespViewer::SwitchSelectMode()
{
	if (selectMode == SELECT_VERTICES) {
		selectMode = SELECT_CORRESP;
	}
	else if (selectMode == SELECT_CORRESP) {
		selectMode = SELECT_VERTICES;
	}
}

void CorrespViewer::CaptureResults( const std::string& name, const std::string& camFile )
{
	if (camFile.length()) {
		// Load the camera view
		sourceView->loadCamera(QString(camFile.c_str()));
	}
	sourceView->showModelsAct->setChecked(false);
	sourceView->showMotionSegmentationAct->setChecked(true);
	targetView->showMotionSegmentationAct->setChecked(true);
	sourceView->resetView();

	// Save a snapshot
	sourceView->saveScreenshot2(QString("%1_src.png").arg(name.c_str()));
	targetView->saveScreenshot2(QString("%1_tgt.png").arg(name.c_str()));
}
