#include "DeformEdit.h"
#include <iostream>
#include "viewerCommon.h"

DeformEdit::DeformEdit(QWidget *parent)
	: QWidget(parent)
{
	ui.setupUi(this);
	setAttribute(Qt::WA_DeleteOnClose);		// Destroy the object when closing

	init();
}

DeformEdit::~DeformEdit()
{
	for (int i = 0; i < controls.size(); i++) {
		delete controls[i];
	}
}

void DeformEdit::init()
{
	controls.clear();

	// IMPORTANT!
	// To add a new control, insert the line here!  That's it!
	controls.push_back(new ParamSelect(this, "# Bones", NUM_BONES));
	controls.push_back(new ParamSelect(this, "# Grid Divisions", NUM_DIVISIONS));
	controls.push_back(new ParamSelect(this, "# Correspondences", NUM_CORRESPONDENCES));
	controls.push_back(new ParamSelect(this, "Max Corresp Dist", MAX_CORRESP_DIST));
	controls.push_back(new ParamSelect(this, "Max Normal Difference", MAX_NORMAL_DIFFERENCE));
	controls.push_back(new ParamSelect(this, "Max Bdy Corresp Dist", MAX_BDY_CORRESP_DIST));
	controls.push_back(new ParamSelect(this, "E-step Constraint Weight", DEFORM_EM_CONSTRAINT_E_WEIGHT));
	controls.push_back(new ParamSelect(this, "E-step Smoothness Weight", DEFORM_EM_SMOOTHNESS_E_WEIGHT));
	controls.push_back(new ParamSelect(this, "Solver # Iterations", SOLVER_MAX_ITERATIONS));
	controls.push_back(new ParamSelect(this, "E-step # Iterations", ESTEP_MAX_ITERATIONS));
	controls.push_back(new ParamSelect(this, "Colormap Min", DEFORM_COLORMAP_RANGE_MIN));
	controls.push_back(new ParamSelect(this, "Colormap Max", DEFORM_COLORMAP_RANGE_MAX));

	//controls.push_back(new ParamSelect(this, "M-step Constraint Weight", DEFORM_EM_CONSTRAINT_M_WEIGHT));
	//controls.push_back(new ParamSelect(this, "M-step Smoothness Weight", DEFORM_EM_SMOOTHNESS_M_WEIGHT));
	//controls.push_back(new ParamSelect(this, "M-step Unity Weight", DEFORM_EM_UNITY_M_WEIGHT));
	//controls.push_back(new ParamSelect(this, "M-step Discrete Weight", DEFORM_EM_DISCRETE_M_WEIGHT));
	//controls.push_back(new ParamSelect(this, "Proj # Iterations", PROJ_MAX_ITERATIONS));
	//controls.push_back(new ParamSelect(this, "IRLS # Iterations", IRLS_MAX_ITERATIONS));

	//////////////////////////////////////////////////////////////////////////
	for (int i = 0; i < controls.size(); i++) {
		connect(controls[i], SIGNAL(valueChanged(int, double)), this, SLOT(controlChanged(int, double)));
	}
}

void DeformEdit::controlChanged(int id, double value)
{
	emit valueChanged(id, value);
}

ParamSelect* DeformEdit::findControl(int id)
{
	for (int i = 0; i < controls.size(); i++) {
		if (controls[i]->getId() == id) {
			return controls[i];
		}
	}

	return NULL;
}

void DeformEdit::initRange( int which, double minRange, double maxRange, double initValue )
{
	ParamSelect *control;
	if (control = findControl(which)) {
		control->initRange(minRange, maxRange, initValue);
	}
	else {
		std::cerr << "Warning in DeformEdit::initRange(): Control " << which << " not found!" << std::endl;
	}
}

void DeformEdit::on_outputStats_clicked()
{
	emit valueChanged(OUTPUT_STATS, 1);
}

void DeformEdit::on_setupDeformButton_clicked()
{
	emit valueChanged(SETUP_DEFORM, 1);
}

void DeformEdit::on_solveDeformInitButton_clicked()
{
	emit valueChanged(SOLVE_DEFORM_INIT, 1);
}

void DeformEdit::on_solveDeformTransformButton_clicked()
{
	emit valueChanged(SOLVE_DEFORM_TRANSFORM, 1);
}

void DeformEdit::on_solveDeformWeightButton_clicked()
{
	emit valueChanged(SOLVE_DEFORM_WEIGHT, 1);
}

void DeformEdit::on_solveDeformUpdateCorrespButton_clicked()
{
	emit valueChanged(SOLVE_DEFORM_CORRESP, 1);
}

void DeformEdit::on_solveDeformAutoButton_clicked()
{
	emit valueChanged(SOLVE_DEFORM_AUTO, 1);
}

void DeformEdit::on_exportResultButton_clicked()
{
	emit valueChanged(EXPORT_DEFORM_RESULT, 1);
}

void DeformEdit::on_inverseDeformButton_clicked()
{
	emit valueChanged(INVERSE_DEFORM, 1);
}

void DeformEdit::on_stepSolverButton_clicked()
{
	emit valueChanged(STEP_SOLVER, 1);
}

void DeformEdit::on_pauseSolverButton_clicked()
{
	emit valueChanged(PAUSE_SOLVER, 1);
}

void DeformEdit::on_stopSolverButton_clicked()
{
	emit valueChanged(STOP_SOLVER, 1);
}

void DeformEdit::on_OtherTasksComboBox_currentIndexChanged(int index)
{
	if (index > 0) {
		emit valueChanged(OTHER_TASKS, (double) index);
		ui.OtherTasksComboBox->setCurrentIndex(0);
	}
}

void DeformEdit::on_solveInitialRigidCheckBox_stateChanged(int status)
{
	emit valueChanged(SOLVE_INITIAL_RIGID, status);
}

void DeformEdit::on_correspondenceTypeCombo_currentIndexChanged(int index)
{
	emit valueChanged(UPDATE_CORRESPONDENCE_TYPE, index);
}

void DeformEdit::init_solveInitialRigidCheckBox( bool value )
{
	ui.solveInitialRigidCheckBox->setChecked(value);
}

void DeformEdit::init_correspondenceTypeCombo( int type )
{
	ui.correspondenceTypeCombo->setCurrentIndex(type);
}

void DeformEdit::on_AccumulateGeometryButton_clicked()
{
	emit valueChanged(ACCUMULATE_GEOMETRY, 1);
}

void DeformEdit::on_requestNextFrameButton_clicked()
{
	emit valueChanged(REQUEST_NEXT_FRAME, 1);
}

void DeformEdit::on_RevertSolutionButton_clicked()
{
	emit valueChanged(REVERT_PREVIOUS_SOLUTION, 1);
}

void DeformEdit::on_solveFinalWeightsCheckBox_stateChanged(int status)
{
	emit valueChanged(SOLVE_FINAL_WEIGHTS, status);

}

void DeformEdit::init_solveFinalWeightsCheckBox( bool value )
{
	ui.solveFinalWeightsCheckBox->setChecked(value);
}

