// From the multiSelect example
#include <QMenu>
#include <QGLViewer/qglviewer.h>
#include "types.h"

class Scene;
class ShapeModel;
class SE3;
class PointList;

// Compute registration error
class ErrorStats
{
public:
	ErrorStats() : maxError(-REAL_TYPE_MAX), minError(REAL_TYPE_MAX),
		avgError(0.f), varError(0.f), numValid(0) { }

	void print( real_type depthScan_gridSpacing, size_t numVertices );
	void printCSV( size_t numVertices );

	real_type maxError;
	real_type minError;
	real_type avgError;
	real_type varError;
	int numValid;
};

class TransformViewer : public QGLViewer
{
public:
	TransformViewer( Scene *sc );
	virtual ~TransformViewer();

protected :
	virtual void init();
	virtual void draw();
	virtual void keyPressEvent(QKeyEvent *e);
	virtual void dragEnterEvent(QDragEnterEvent *event);
	virtual void dropEvent(QDropEvent *event);

private :
	void batchComputeError();
	void computeRegistrationError();
	void drawRegistrationError();
	void saveRegistrationErrorBinary( const char *savefileName );
	void saveRegistrationErrorText( QString &savefileName );
	void drawColormap();
	ShapeModel* getCompareShape_Source()
	{
		return (useFrameSource < 0 || useFrameSource >= sourceFrames.size()) ? NULL : sourceFrames[useFrameSource];
	}
	ShapeModel* getCompareShape_Target()
	{
		return (useFrameTarget < 0 || useFrameTarget >= targetFrames.size()) ? NULL : targetFrames[useFrameTarget];
	}
	distList* getRegistrationError()
	{
		return (useFrameError < 0 || useFrameError >= errorFrames.size()) ? NULL : &(errorFrames[useFrameError]);
	}
	PointList* getPts()
	{
		return (useFramePts < 0 || useFramePts >= ptsFrames.size()) ? NULL : ptsFrames[useFramePts];
	}
	void processDropFiles( QStringList &fileList, const Qt::KeyboardModifiers modifiers );
	void processObjSequence( QStringList &fileList, bool loadTargetFrames );

	// Shapes
	Scene *scene;
	bool showSource, showTarget;
	bool useSameColor;
	real_type depthScan_gridSpacing;

	ErrorStats registrationStats;
	distList registrationError;
	bool showRegistrationError;
	const real_type INVALID_ERROR_MARK;			/**< Need a value to mark invalid comparisons due to vertices on boundary */


	// Drawing
	bool showWireframe;
	QColor invalidRegistrationColor;

	//////////////////////////////////////////////////////////////////////////
	// Additional frames loaded
	int useFrameSource;
	std::vector<ShapeModel*> sourceFrames;

	int useFrameTarget;
	std::vector<ShapeModel*> targetFrames;

	int useFrameError;
	std::vector<distList> errorFrames;

	int useFramePts;
	std::vector<PointList*> ptsFrames;
	std::vector<PointList*> normalsFrames;

	//////////////////////////////////////////////////////////////////////////
	// Menu variables and functions
	void createContextMenu();

	QMenu* contextMenu;
	QAction* loadShapeAct;
};
