#include "viewer.h"
#include "scene.h"
#include "viewerCommon.h"
#include "CorrespViewer.h"
#include "DeformEdit.h"
#include "ShapePoser.h"
#include "TransformViewer.h"

using namespace std;

void Viewer::CreateDeformEditor()
{
	if (!deformEdit) {

		bool ok = true;
		// Check that we have enough weights
		if (scene->em_EWeights.size() < 2) {
			cerr << "Not enough E-step weights (only have " << scene->em_EWeights.size() << ") for DeformEM." << endl;
			ok = false;
		}
		if (scene->em_DiscreteMWeights.size() < 2) {
			cerr << "Not enough Discrete M-step weights (only have " << scene->em_DiscreteMWeights.size() << ")for DeformEM." << endl;
			ok = false;
		}
		//if (scene->em_MWeights.size() < 4) {
		//	cerr << "Not enough M-step weights (only have " << scene->em_MWeights.size() << ")for DeformEM." << endl;
		//	ok = false;
		//}

		if (!ok) return;

		deformEdit = new DeformEdit();
		deformEdit->initRange(NUM_BONES, 0, scene->graph_n*2, scene->graph_n);
		deformEdit->initRange(NUM_DIVISIONS, 0, scene->gridDivisions*2, scene->gridDivisions);
		deformEdit->initRange(NUM_CORRESPONDENCES, 0, scene->numCorrespondences*2, scene->numCorrespondences);
		deformEdit->initRange(MAX_CORRESP_DIST, 0, scene->maxCorrespondenceDistance*2, scene->maxCorrespondenceDistance);
		deformEdit->initRange(MAX_NORMAL_DIFFERENCE, 0, scene->maxNormalAngleDifference*2, scene->maxNormalAngleDifference);
		deformEdit->initRange(MAX_BDY_CORRESP_DIST, 0, scene->boundaryCorrespondenceDistance*2, scene->boundaryCorrespondenceDistance);
		deformEdit->initRange(DEFORM_EM_CONSTRAINT_E_WEIGHT, 0, scene->em_EWeights[0]*2, scene->em_EWeights[0]);
		deformEdit->initRange(DEFORM_EM_SMOOTHNESS_E_WEIGHT, 0, scene->em_EWeights[1]*2, scene->em_EWeights[1]);
		deformEdit->initRange(SOLVER_MAX_ITERATIONS, 0, scene->solverMaxIterations*2, scene->solverMaxIterations);
		deformEdit->initRange(ESTEP_MAX_ITERATIONS, 0, scene->estepMaxIterations*2, scene->estepMaxIterations);
		deformEdit->initRange(DEFORM_COLORMAP_RANGE_MIN, 0, scene->colormapMin*2, scene->colormapMin);
		deformEdit->initRange(DEFORM_COLORMAP_RANGE_MAX, 0, scene->colormapMax*2, scene->colormapMax);

		//deformEdit->initRange(DEFORM_EM_CONSTRAINT_M_WEIGHT, 0, scene->em_MWeights[0]*2, scene->em_MWeights[0]);
		//deformEdit->initRange(DEFORM_EM_SMOOTHNESS_M_WEIGHT, 0, scene->em_MWeights[1]*2, scene->em_MWeights[1]);
		//deformEdit->initRange(DEFORM_EM_UNITY_M_WEIGHT, 0, scene->em_MWeights[2]*2, scene->em_MWeights[2]);
		//deformEdit->initRange(DEFORM_EM_DISCRETE_M_WEIGHT, 0, scene->em_MWeights[3]*2, scene->em_MWeights[3]);
		//deformEdit->initRange(PROJ_MAX_ITERATIONS, 0, scene->projMaxIterations*2, scene->projMaxIterations);
		//deformEdit->initRange(IRLS_MAX_ITERATIONS, 0, scene->IRLSMaxIterations*2, scene->IRLSMaxIterations);
		deformEdit->init_solveInitialRigidCheckBox(scene->solveInitialRigid);
		deformEdit->init_solveFinalWeightsCheckBox(scene->solveFinalNNLS);
		deformEdit->init_correspondenceTypeCombo(scene->correspondenceType);

		// Show the UI dialog
		deformEdit->show();

		connect(deformEdit, SIGNAL(valueChanged(int,double)), this, SLOT(DeformEditorChanged(int,double)));
		connect(deformEdit, SIGNAL(destroyed()), this, SLOT(DeformEditorDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		deformEdit->hide();
		deformEdit->show();
	}
}

void Viewer::DeformEditorDestroyed()
{
	//cerr << "Deformation control is: " << deformEdit << endl;
	deformEdit = NULL;

	// Print the values of the parameters for verification
	cout << "----------------------------------------------" << endl;
	cout << "Final Parameter Values:" << endl;
	cout << "----------------------------------------------" << endl;
	cout << "Graph Nodes: " << scene->graph_n << endl;
	cout << "Graph Neighbors: " << scene->graph_k << endl;
	cout << "Solver Max Iterations: " << scene->solverMaxIterations << endl;
	cout << "Projection Max Iterations: " << scene->projMaxIterations << endl;
	cout << "----------------------------------------------" << endl;
}

void Viewer::DeformEditorChanged(int which, double value)
{
	QString message;

	switch(which) {
	case OUTPUT_STATS:
		// Nothing implemented yet!
		break;
	case OTHER_TASKS:
		switch ((int)value) {
			//case 1:
			//	CaptureResultSnapshots_AlignPair();
			//	break;
			//case 2:
			//	CaptureResultSnapshots_AlignSeq();
			//	break;
			//case 3:
			//	CaptureResultSnapshots_GridWeights();
			//	break;
			//case 4:
			//	scene->printCurrentDeformation();
			//	break;
			//case 5:
			//	{
			//		QString saveFilename = QFileDialog::getSaveFileName(this, QString("Save Registration Result"), scene->lastDir, QString("Data (*.data)"));
			//		scene->saveDeformation(saveFilename);
			//	}
			//	break;
			//case 6:
			//	{
			//		QString loadFilename = QFileDialog::getOpenFileName(this, QString("Load Registration Result"), scene->lastDir, QString("Data (*.data)"));
			//		scene->loadDeformation(loadFilename);
			//	}
			//	break;
	default:
		break;
		}
		break;
	case NUM_BONES:
		scene->graph_n = value;
		cout << "Setting # bones: " << scene->graph_n << endl;
		break;
	case NUM_DIVISIONS:
		scene->gridDivisions = value;
		cout << "Setting # grid divisions on longest axis of bbox: " << scene->gridDivisions << endl;
		break;
	case NUM_CORRESPONDENCES:
		scene->numCorrespondences = value;
		cout << "Setting # correspondences: " << scene->numCorrespondences << endl;
		break;
	case MAX_CORRESP_DIST:
		scene->maxCorrespondenceDistance = value;
		cout << "Setting max correspondence distance: " << scene->maxCorrespondenceDistance << endl;
		scene->setDeformOptions();
		break;
	case MAX_NORMAL_DIFFERENCE:
		scene->maxNormalAngleDifference = value;
		cout << "Setting max normal angle difference: " << scene->maxNormalAngleDifference << endl;
		scene->setDeformOptions();
		break;
	case MAX_BDY_CORRESP_DIST:
		scene->boundaryCorrespondenceDistance = value;
		cout << "Setting max boundary correspondence distance: " << scene->boundaryCorrespondenceDistance << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_CONSTRAINT_E_WEIGHT:
		scene->em_EWeights[0] = value;
		cout << "Setting DeformEM E-step constraint weight: " << scene->em_EWeights[0] << endl;
		//scene->em_multiframeEWeights[0] = value;
		//cout << "Setting DeformEM E-step constraint weight: " << scene->em_multiframeEWeights[0] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_SMOOTHNESS_E_WEIGHT:
		scene->em_EWeights[1] = value;
		cout << "Setting DeformEM E-step constraint weight: " << scene->em_EWeights[1] << endl;
		//scene->em_multiframeEWeights[1] = value;
		//cout << "Setting DeformEM E-step smoothness weight: " << scene->em_multiframeEWeights[1] << endl;
		scene->setDeformOptions();
		if (value == 0.0f) {
			displayMessage(tr("Using NO Smoothness!"));
		}
		else {
			displayMessage(tr("Setting Smoothness term weight: %1").arg(value));
		}
		break;
	case DEFORM_EM_CONSTRAINT_M_WEIGHT:
		scene->em_MWeights[0] = value;
		cout << "Setting DeformEM M-step constraint weight: " << scene->em_MWeights[0] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_SMOOTHNESS_M_WEIGHT:
		scene->em_MWeights[1] = value;
		cout << "Setting DeformEM M-step smoothness weight: " << scene->em_MWeights[1] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_UNITY_M_WEIGHT:
		scene->em_MWeights[2] = value;
		cout << "Setting DeformEM M-step unity weight: " << scene->em_MWeights[2] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_DISCRETE_M_WEIGHT:
		scene->em_MWeights[3] = value;
		cout << "Setting DeformEM M-step discrete weight: " << scene->em_MWeights[3] << endl;
		scene->setDeformOptions();
		break;
	case SOLVER_MAX_ITERATIONS:
		scene->solverMaxIterations = value;
		cout << "Setting Solver Max Iterations: " << scene->solverMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case ESTEP_MAX_ITERATIONS:
		scene->estepMaxIterations = value;
		cout << "Setting E-Step Max Iterations: " << scene->estepMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case PROJ_MAX_ITERATIONS:
		scene->projMaxIterations = value;
		cout << "Setting E-Step Projection Max Iterations: " << scene->projMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case IRLS_MAX_ITERATIONS:
		scene->IRLSMaxIterations = value;
		cout << "Setting M-Step IRLS Max Iterations: " << scene->IRLSMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case SOLVE_INITIAL_RIGID:
		scene->solveInitialRigid = value;
		cout << "Setting solve initial rigid transformation: " << (scene->solveInitialRigid?"Yes":"No") << endl;
		scene->setDeformOptions();
		break;
	case SOLVE_FINAL_WEIGHTS:
		scene->solveFinalNNLS = value;
		cout << "Setting solve final continuous weights: " << (scene->solveFinalNNLS?"Yes":"No") << endl;
		scene->setDeformOptions();
		break;
	case UPDATE_CORRESPONDENCE_TYPE:
		scene->correspondenceType = value;
		cout << "Setting correspondence update type: " << scene->correspondenceType << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_COLORMAP_RANGE_MIN:
		scene->colormapMin = value;
		cout << "Setting colormap min: " << scene->colormapMin << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_COLORMAP_RANGE_MAX:
		scene->colormapMax = value;
		cout << "Setting colormap max: " << scene->colormapMax << endl;
		scene->setDeformOptions();
		break;
	case EXPORT_DEFORM_RESULT:
		// Export the deformed mesh
		if (scene->is_deformation_set_up()) {
			cerr << "Not implemented yet!" << endl;
			//ShapeModel *shape = scene->deformResult();

			//if (shape->shapeType == SHAPE_MESH) {
			//	// Prompt for a save file
			//	QString savefileName;
			//	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the deformed mesh...",
			//												scene->lastDir, QString("*.obj"));
			//	if (savefileName.length() > 0) {
			//		((Mesh3_real*)shape)->writeGeometryOBJ(qPrintable(savefileName));
			//	}
			//}
		}
		break;
	case INVERSE_DEFORM:
		// Functionality not supported anymore
		cerr << "This functionality is not supported anymore!" << endl;

		////// Perform inverse deformation
		////if (scene->is_deformation_set_up()) {
		////	ShapeModel *shape = scene->inverseDeform(*(scene->target));

		////	// Add the result to the list of meshes
		////	scene->getAnimation()->add(shape);
		////}
		break;

		// Auto-solver thread control
	case SETUP_DEFORM:
		scene->controlSolveThread(which, whichDeform());
		fitViewToShape();
		updateGL();
		break;
	case SOLVE_DEFORM_AUTO:
	case PAUSE_SOLVER:
	case STEP_SOLVER:
	case STOP_SOLVER:
		scene->controlSolveThread(which, whichDeform());
		updateGL();
		break;

	case REVERT_PREVIOUS_SOLUTION:
		//	// Revert to the discrete weights
		//	scene->revertDiscreteWeights();
		//	updateGL();
		//	break;
	case SOLVE_DEFORM_WEIGHT:
		//	// Solve final NNLS weights
		//	scene->solveFinalNNLSWeights();
		//	updateGL();
		//	break;

		// Fine solver control: execute any step of the solver that you want
	case ACCUMULATE_GEOMETRY:
	case REQUEST_NEXT_FRAME:
	case SOLVE_DEFORM_INIT:
	case SOLVE_DEFORM_TRANSFORM:
	case SOLVE_DEFORM_CORRESP:
		cerr << "This functionality is not implemented yet!" << endl;
		break;

		////// Old stuff
		////case GRAPH_NODES:
		////	scene->graph_n = maxGraphNodes * value;
		////	deformEdit->displayValue(which, tr("%1").arg(scene->graph_n));
		////	scene->resampleSource();
		////	scene->setupDeform(whichDeform());
		////	break;
		////case GRAPH_NEIGHBORS:
		////	scene->graph_k = maxGraphNeighbors * value;
		////	deformEdit->displayValue(which, tr("%1").arg(scene->graph_k));
		////	scene->setupDeform(whichDeform());
		////	break;
		////case OVERLAP_FACTOR:
		////	scene->overlapf = maxOverlapFactor * value;
		////	deformEdit->displayValue(which, tr("%1").arg(scene->overlapf));
		////	scene->setupDeform(whichDeform());
		////	break;
		////case MAX_ITERATIONS_SOLVER:
		////	scene->solverMaxIterations = value;
		////	break;
		////case MAX_ITERATIONS_PROJECTION:
		////	scene->projMaxIterations = value;
		////	break;
		////case ED_WEIGHTS_ROTATION:
		////	scene->edWeights[0] = value;
		////	break;
		////case ED_WEIGHTS_SMOOTH:
		////	scene->edWeights[1] = value;
		////	break;
		////case ED_WEIGHTS_CONSTRAINT:
		////	scene->edWeights[2] = value;
		////	break;
		////case WD_WEIGHTS_SMOOTH:
		////	scene->wdWeights[0] = value;
		////	break;
		////case WD_WEIGHTS_CONSTRAINT:
		////	scene->wdWeights[1] = value;
		////	break;
	default:
		break;
	}

	updateGL();
}

void Viewer::CreateCorrespViewer()
{
	if (!correspViewer) {
		correspViewer = new CorrespViewer(scene, this);
		correspViewer->show();

		//if (featureViewer) {
		//	featureViewer->correspViewer = correspViewer;
		//}

		connect(this, SIGNAL(updateChildDisplay()), correspViewer, SLOT(updateDisplay()));
		connect(correspViewer, SIGNAL(destroyed()), this, SLOT(CorrespViewerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		correspViewer->hide();
		correspViewer->show();
	}
}

void Viewer::CorrespViewerDestroyed()
{
	//cerr << "Correspondence Viewer is: " << correspViewer << endl;
	correspViewer = NULL;

	//if (featureViewer) {
	//	featureViewer->correspViewer = correspViewer;
	//}
}

void Viewer::CreateShapePoser()
{
	ShapeModel *shape = scene->source;

	if (!shape) {
		if (scene->is_deformation_set_up(Deform::DEFORM_GRAPH_ALL)) {
			// Assume that the final shape is already loaded -- get it
			shape = scene->getFinalShape();
		}
		else {
			// Open a shape file
			QString qfilename = QFileDialog::getOpenFileName(this, tr("Open OBJ"), scene->lastDir, tr("Meshes (*.obj)"));
			if (qfilename.isEmpty()) {
				return;
			}

			shape = loadShape(tr(scene->rootDir.c_str()), qfilename, *scene, PROCESS_NONE);
		}
	}

	if (shape && !shapePoser) {
		shapePoser = new ShapePoser(scene, shape);

		// subsample the constraints for faster performance
		shapePoser->subsampleConstraints(scene->numCorrespondences);

		shapePoser->show();

		// Resizing for video
		shapePoser->resize(800, 600);

		connect(this, SIGNAL(updateChildDisplay()), shapePoser, SLOT(updateGL()));
		connect(shapePoser, SIGNAL(destroyed()), this, SLOT(ShapePoserDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		shapePoser->hide();
		shapePoser->show();
	}
}

void Viewer::ShapePoserDestroyed() {
	//cerr << "Shape Poser is: " << shapePoser << endl;
	scene->clearShapePoserInstance();
	shapePoser = NULL;
}


void Viewer::CreateTransformViewer()
{
	ShapeModel *shape = selectCurrentModel();

	if (!transformViewer) {
		transformViewer = new TransformViewer(scene);
		transformViewer->show();

		connect(this, SIGNAL(updateChildDisplay()), transformViewer, SLOT(updateGL()));
		connect(transformViewer, SIGNAL(destroyed()), this, SLOT(TransformViewerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		transformViewer->hide();
		transformViewer->show();
	}
}

void Viewer::TransformViewerDestroyed() {
	//cerr << "TransformViewer is: " << transformViewer << endl;
	transformViewer = NULL;
}


bool Viewer::captureCorrespViewerResult( int winH, int winV )
{
	if (correspViewer) {
		correspViewer->resize(winH, winV);
		correspViewer->CaptureResults(scene->autoSaveCorrespondenceResult, scene->autoLoadCameraParams);
		return true;
	}
	return false;
}

void Viewer::correspViewerPosition( int x, int y, int width /*= 0*/, int height /*= 0 */ )
{
	setViewerPosition(static_cast<QWidget&>(*correspViewer), x, y, width, height);
}

void Viewer::deformEditPosition( int x, int y, int width /*= 0*/, int height /*= 0 */ )
{
	setViewerPosition(static_cast<QWidget&>(*deformEdit), x, y, width, height);
}
